from copy import deepcopy
from typing import Union
from cli_args_system.list_args import ListArgs


# The Flags Content Extends ListArgs, witch is a "only-read" list
class FlagsContent(ListArgs):
    
    def __init__(self, content: Union[None, list]) -> None:
        """content:the group of args that were found \n"""

        super().__init__()
        # set the exist and filled to false
        self._exist = False
        self._filled = False
        # if content is not None, means the flag exist
        if content is not None:
            self._exist = True
            # if content is not [], means its filled
            if content:
                self._filled = True
                # set the args to content, for to be
            # used with the super(ListArgs)
            self._args = content


    def flags(self)->list:
        """returns The founded flags"""
        return deepcopy(self._args) 


    def exist(self)->bool:
        """returns True if flag were passed on argv else False"""
        return self._exist
    
    def exist_and_empty(self)->bool:
        """returns True if exist and is not filled"""
        return True if self._exist and not self._filled else False 

    def filled(self)->bool:
        """returns True if flag is filled else False, ex: \n
        flags = [a.txt, b.txt ]-> True \n
        flags = [] -> False""" 
        return self._filled
    
    def __repr__(self) -> str:
        return f'exist:  {self._exist}\n'\
               f'filled: {self._filled}\n'\
               f'args:   {self._args}' 
                


from typing import Union
from cli_args_system.extras import get_num_comparison


# abstract class
# the ListArgs is a "only read" list, that implements the
# "magic" methods of lists
class ListArgs:

    def __init__(self) -> None:
        # creates the empty args
        self._args = []
    
    def __eq__(self, o: object) -> bool:
        """this method is called when == is called
        o: the object comparison, ex args == 20, 20 its "o"
        """
        # if comparison is a int, it will be compared with size
        if isinstance(o, int):
            return len(self) == o
        # if its a list it will be compared with the args property
        if isinstance(o, list):
            return o == self._args

        if isinstance(o, tuple):
            return list(o) == self._args

        return False

    def __ne__(self, o: object) -> bool:
        """this method is called when != is called"""
        # its "flip" the __eq__ methods
        return False if self == o else True

    def __getitem__(self, index: Union[slice, int]):
        """this method its called when != is used """
        # its just return the index element
        return self._args[index]

    def __contains__(self, o: object):
        """this method is called when "in" is used"""
        # its just return the in comparison with args
        return o in self._args

    def __len__(self):
        """this method is called when len(x) is called"""
        return len(self._args)

    # all above methods is for >, >=, <. <= methods
    # and all make the comparison with the self len
    def __gt__(self, o: Union[int, list]):
        return len(self) > get_num_comparison(o)

    def __ge__(self, o: Union[int, list]):
        return len(self) >= get_num_comparison(o)

    def __lt__(self, o: Union[int, list]):
        return len(self) < get_num_comparison(o)

    def __le__(self, o: Union[int, list]):
        return len(self) <= get_num_comparison(o)



from typing import Union


def format_flag(
        flag: str,
        case_sensitive: bool,
        flag_identifier: str,
        infinity_identifier: bool) -> str:
    """this method is responsible for format a given flag \n
    flag: the flag \n
    case_sensitive: if false all will be lower\n
    flag_identifier: the char who identifies the flag ex: -b i=(-)\n
    infinity_identifier: if its possible to set ---x flags\n"""
    if not case_sensitive:
        flag = flag.lower()

    # if identifier is false, just remove the first char
    if not infinity_identifier:
        return flag[1::]

    # this loop is to find the first char that
    # is not the identifier
    for index, char in enumerate(flag):
        # when find the difference split the flag with the
        # current index
        if char != flag_identifier:
            return flag[index::]

    return flag


def is_a_flag(
        possible_flag: Union[str, int, float],
        case_sensitive: bool,
        flag_identifier: str) -> bool:
    """this method is used to recognize flags\n
    case_sensitive: if not wil compare in lower_case \n
    flag_identifier: the identifier tha tels if its a flag\n"""

    # this is for when the method is called
    # with numbers args
    if possible_flag.__class__ in [int, float]:
        return False

    if not case_sensitive:
        possible_flag = possible_flag.lower()

    # if char[0] its equal to flag_identifier and the
    # len of comparison is > 1 means its a flag
    # the len comparison is to not return true for "-" args
    if possible_flag[0] == flag_identifier and len(possible_flag) > 1:
        return True

    return False


def get_flags(
        args: list,
        flag_identifier: str,
        case_sensitive: bool,
        infinity_identifier: bool) -> dict:
    """this method returns a dict, with all flags passed in args \n
    args: the list of args \n
    case_sensitive: if is False, the flags will be lower_case
    flag_identifier: the char that identifiers what is a flag, ex:"-"\n
    infinity_identifier: is is False, just the first element will be
    considered as a flag identifier \n"""
    # create the default dict, and default flag
    flags: dict = {'default': []}
    current_flag = 'default'

    # go in a loop for all args
    for arg in args:

        # if is a flag is true, changes the current flags
        if is_a_flag(possible_flag=arg, case_sensitive=case_sensitive, flag_identifier=flag_identifier):
            # changes the current flag
            current_flag = format_flag(
                flag=arg,
                case_sensitive=case_sensitive,
                flag_identifier=flag_identifier,
                infinity_identifier=infinity_identifier
            )
            # this is to prevent previews equals current flags
            if current_flag not in flags.keys():
                # creates a new list inside the flags dict
                flags[current_flag] = []
        else:
            # if is not a flag, append with tue current flag
            flags[current_flag].append(arg)
    return flags


def convert_number(possible_number: str) -> Union[str, float, int]:
    """convert the possible_number  in a number if its possible,
    case if were not possible to convert, will return the possible_number
    (the same arg of method)"""

    try:
        # try to cast in an int
        return int(possible_number)
    except ValueError:
        try:
            # case cat a Value error, try to cast in a float
            return float(possible_number)
        except ValueError:
            # case if were not a int either a float, will cat here
            # and return the same arg
            return possible_number


def format_args(args: list, consider_first: bool, case_sensitive=bool, convert_numbers=bool):
    """this method format args, based on args passed
    args: the list of "argv"
    consider_first: if the first element of args is to be considered
    case_sensitive: if False, all will be lower
    convert_numbers: convert numbers in float or int if possible"""
    # split the args removing the first occurrence, if not consider first
    if not consider_first:
        args = args[1::]

    # map the args to all be lower
    if not case_sensitive:
        args = list(map(lambda arg: arg.lower(), args))

    # make a mapping calling the convert number function
    if convert_numbers:
        args = list(map(convert_number, args))
    return args


def cast_list(*elements) -> list:
    """this method transform everything in a list, ex:\n
    cast_list("a","b") returns: ["a","b"]\n
    cast_list([1,2]) returns: [1,2]\n"""
    # when * is used all the arguments will be passed as tuple

    # means just one element were passed as arg
    if len(elements) == 1:
        # means nothing were passed as arg
        if elements[0] is None:
            return []

        # means a tuple or list were passed as argument
        if elements[0].__class__ in [list, tuple]:
            # transform this in a list
            return list(elements[0])

        # else will make a list with elements 0
        # this is when a str or int is passed
        return [elements[0]]
    else:

        # if its a larger element, just cast with list
        return list(elements)


def get_num_comparison(o: Union[int, list, tuple]) -> Union[int, None]:
    object_type = o.__class__
    if object_type == int:
        return o
    if object_type in [list, tuple]:
        return len(o)
    return None



from sys import argv
from copy import deepcopy
from json import dumps
from typing import Any, Union

from cli_args_system.extras import format_args, get_flags, cast_list
from cli_args_system.flags_content import FlagsContent
from cli_args_system.list_args import ListArgs


class Args(ListArgs):

    def __init__(self, consider_first=False, flags_case_sensitive=False, args_case_sensitive=True, convert_numbers=True,
                 flag_identifier='-', infinity_identifier=True, args=None) -> None:
        """consider_first: if is to consider the first argv element,
        if false, argv[0] will be removed\n
        flags_case_sensitive: if false, all flags will be lowercase \n
        convert_numbers: if True, it will convert in float or int all
        found numbers in argv\n
        flag_identifier: the char that is used to identifies a flag, ex:
        in "--a", the flag_identifier is "-"\n
        infinity_identifier: if false, just the first char will be considered
        as the flag identifier\n
        """
        super().__init__()
        if args is None:
            args = deepcopy(argv)
        self._args = format_args(
            args=args,
            consider_first=consider_first,
            case_sensitive=args_case_sensitive,
            convert_numbers=convert_numbers
        )
        
        
        self._flags = get_flags(
            args=self._args,
            flag_identifier=flag_identifier,
            case_sensitive=flags_case_sensitive,
            infinity_identifier=infinity_identifier
        )
        self._unused_flags = deepcopy(self._flags)


    def args(self) ->list:
        """returns the arguments passed in argv, 
        formatted according to the initial arguments of the class"""
        return deepcopy(self._args)

    
    def flags_dict(self)->dict:
        """returns a dictionary of flags captured in argv"""
        return  deepcopy(self._flags) 


    def flags_names(self,include_default=False) ->list:
        """returns the name of the flags passed in argv\n
        include_default: if False, the default flags will not
        be included on list
        """
        flags = list(self._flags.keys())
        return flags if include_default else flags[1::]


    def total_flags(self,include_default=False) ->int:
        """returns the total size of flags\n
        include_default: if False, the default flags will not
        be included on list"""
        return len(self.flags_names(include_default))


    def unused_flags(self)->dict:
        """returns the flags that were not used after 
        all the flags_content were called
        """
        return  deepcopy(self._unused_flags) 


    def unused_flags_names(self,include_default=False)->list:
        """returns the flags names that were not used after 
        all the flags_content were called\n
        include_default: if False, the default flags will not
        be included on list"""
        flags = list(self._unused_flags.keys())
        return flags if include_default else flags[1::]


    def total_unused_flags(self,include_default=False)->int:
        """returns the total size of flags that were not used after 
        all the flags_content were called\n
        include_default: if False, the default flags will not
        be included on list"""
        return len(self.unused_flags_names(include_default))

    def flag_str(self, *flags)->Union[str,int,None]:
        """return the first founded flag it can be str or int \n
        if does not find any flag, it will return None\n
        flags: the flags that you want to find
        """
        # generate a patronized list of flags
        flags_list = cast_list(*flags)

        for flag in flags_list:
            if flag.__class__ != str:
                raise TypeError('only str are valid for flags')
            try:
                return self._flags[flag][0]
            except (KeyError,IndexError): pass 
        return None 


    def flags_content(self, *flags) -> FlagsContent:
        """returns a FlagsContent object, witch is a group of
        the found flags in argv\n
        flags: the flags that you want to find,
        you can pass a list, tuple, or str"""

        # generate a patronized list of flags
        flags_list = cast_list(*flags)

        # create the filtered args list that will be used
        # to insert the found flags content
        filtered_args = []

        at_least_one_flag_exist = False

        # loop into the flags lis
        for flag in flags_list:
            if flag.__class__ != str:
                raise TypeError('only str are valid for flags')
            try:
                # try to insert into the filtered list
                filtered_args += self._flags[flag]
                # if pass till here, means this flag were inserted in
                # the filtered args
                self._unused_flags.pop(flag)
                at_least_one_flag_exist = True
            except KeyError:
                pass

        # if at least one flag were found, returns FlagsContent object with
        # filtered args
        if at_least_one_flag_exist:
            return FlagsContent(content=filtered_args)
        else:
            # otherwise returns FlagsContent , with None as content
            return FlagsContent(content=None)

    def __eq__(self, o: Any) -> bool:
        """this methods is called when == is used"""
        #if is a dict compare with flags
        if isinstance(o,dict):
            return o == self._flags
        else:
            return super().__eq__(o)

        

        return False

    def __repr__(self) -> str:
        """returns a json representation of the flags attribute"""
        return dumps(self._flags, indent=4)



from distutils.core import setup
setup(
  name = 'cli-args-system',        
  packages = ['cli_args_system'],
  version = '1.0',    
  license='MIT',     
  description = 'A Cli flags libary  to control argv flags and content',   
  author = 'Mateus Moutinho Queiroz',               
  author_email = 'mateusmoutinho01@gmail.com',      
  url = 'https://github.com/mateusmoutinho/python-cli-args.git',  
  download_url = 'https://github.com/mateusmoutinho/python-cli-args/archive/refs/heads/main.zip',    # I explain this later on
  keywords = ['ARGV', 'ClI ARGUMENTS'], 

  classifiers=[
    'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ]
)